(*                         J is for JavaScript:
 *                    A direct-style correspondence
 *             between Algol-like languages and JavaScript
 *                   using first-class continuations
 *
 *            Olivier Danvy, Chung-chieh Shan, and Ian Zerny
 *                            Copyright 2009
 *                  GPLv3, see LICENSE.txt for details
 *
 *                    www.zerny.dk/j-is-for-js.html
 *
 *                         Version: April 2011
 *)

#use "gen.ml";;

(* Calder Mobiles (Danvy, 2006, p. 14) *)
(* A calder mobil is a binary tree where each node has a weight. *)
(* Direct style implementation from the dissertation.
fun bal3 m = callcc (fn k’ => let (* visit : mobile -> int *)
                           fun visit (OBJ n)
                               = n
                             | visit (BAR (n, m1, m2))
                               = let val n1 = visit m1
                                     val n2 = visit m2
                                 in if n1 = n2
                                    then n + n1 + n2
                                    else throw k’ false
                                 end
                       in let val _ = visit m
                          in true
                          end
                       end)
*)

(* We do not have user types so we use an array. *)
(* We represent each node a an index 'i' in an array. Child nodes *)
(* are given as: 2*i+1, 2*i+2. A leaf node has -1 as children *)
(* The implementation is thus, 
procedure calder (T) =
  fail: return false;
  procedure visit (i) = 
    n := T[i];
    if n = -1 then return 0;
    i  := i*2;
    n1 := visit (i + 1)
    n2 := visit (i + 2)
    if n1 = n2
    then return n + n1 + n2;
    goto failure
  end;
  visit 0;
  return true;
end;
print (calder [
   2,
   5, 1,
   2, 2, 4, 4,
   -1, -1, -1, -1, -1, -1, -1, -1]);
*)

let calder = program
  [ defFun (Var "calder") [Var "T"]
      [ defLbl (Lbl "fail") [ cmdReturn (expBool false) ]
      ; defFun (Var "visit") [Var "i"]
	[ defVar (Var "n")
	; defVar (Var "n1")
	; defVar (Var "n2")
	]
	[ cmdAssign (expVar (Var "n")) (expIndex (expVar (Var "T")) (expVar (Var "i")))
	; cmdIfThen (expEq (expVar (Var "n")) (expInt (-1))) (cmdReturn (expInt 0))
	; cmdAssign (expVar (Var "i")) (expMul (expVar (Var "i")) (expInt 2))
	; cmdAssign (expVar (Var "n1")) (expCall (Var "visit") [expAdd (expVar (Var "i")) (expInt 1)])
	; cmdAssign (expVar (Var "n2")) (expCall (Var "visit") [expAdd (expVar (Var "i")) (expInt 2)])
	; cmdIfThen (expEq (expVar (Var "n1")) (expVar (Var "n2")))
                    (cmdReturn (expAdd (expVar (Var "n")) (expAdd (expVar (Var "n1")) (expVar (Var "n2")))))
	; cmdGoto (Lbl "fail")
	]
      ]
      [ cmdExp (expCall (Var "visit") [ expInt 0 ])
      ; cmdReturn (expBool true)
      ]
  ]
  [ cmdPrint (expCall (Var "calder")
		[ expArray
		    [ expInt 2
		    ; expInt 5; expInt 1
		    ; expInt 2; expInt 2; expInt 4; expInt 4
		    ; expInt (-1); expInt (-1); expInt (-1); expInt (-1); expInt (-1); expInt (-1); expInt (-1); expInt (-1)] ])
  ]

let () = print_string (unPrg calder)
