(*                         J is for JavaScript:
 *                    A direct-style correspondence
 *             between Algol-like languages and JavaScript
 *                   using first-class continuations
 *
 *            Olivier Danvy, Chung-chieh Shan, and Ian Zerny
 *                            Copyright 2009
 *                  GPLv3, see LICENSE.txt for details
 *
 *                    www.zerny.dk/j-is-for-js.html
 *
 *                         Version: April 2011
 *)

#use "gen.ml";;

(* Compiled KMP on the pattern "abcabcacab". (Knuth et al., 1977, p. 330) *)

let kmp = program [
  defFun (Var "kmp") [Var "text"]
    [ defVar (Var "k")
    ; defVar (Var "n")
    ; defLbl (Lbl "L0")
             [ cmdAssign (expVar (Var "k")) (expAdd (expVar (Var "k")) (expInt 1))
	     ; cmdGoto (Lbl "L1")
	     ]
    ; defLbl (Lbl "L1")
             [ cmdIfThen (expNotEq (expIndex (expVar (Var "text")) (expVar (Var "k"))) (expChr 'a'))
	                 (cmdGoto (Lbl "L0"))
	     ; cmdAssign (expVar (Var "k")) (expAdd (expVar (Var "k")) (expInt 1))
	     ; cmdIfThen (expGt (expVar (Var "k")) (expVar (Var "n")))
	                 (cmdReturn (expInt (-1)))
	     ; cmdGoto (Lbl "L2")
	     ]
    ; defLbl (Lbl "L2")
             [ cmdIfThen (expNotEq (expIndex (expVar (Var "text")) (expVar (Var "k"))) (expChr 'b'))
	                 (cmdGoto (Lbl "L1"))
	     ; cmdAssign (expVar (Var "k")) (expAdd (expVar (Var "k")) (expInt 1))
	     ; cmdGoto (Lbl "L3")
	     ]
    ; defLbl (Lbl "L3")
             [ cmdIfThen (expNotEq (expIndex (expVar (Var "text")) (expVar (Var "k"))) (expChr 'c'))
	                 (cmdGoto (Lbl "L1"))
	     ; cmdAssign (expVar (Var "k")) (expAdd (expVar (Var "k")) (expInt 1))
	     ; cmdGoto (Lbl "L4")
	     ]
    ; defLbl (Lbl "L4")
             [ cmdIfThen (expNotEq (expIndex (expVar (Var "text")) (expVar (Var "k"))) (expChr 'a'))
	                 (cmdGoto (Lbl "L0"))
	     ; cmdAssign (expVar (Var "k")) (expAdd (expVar (Var "k")) (expInt 1))
	     ; cmdGoto (Lbl "L5")
	     ]
    ; defLbl (Lbl "L5")
             [ cmdIfThen (expNotEq (expIndex (expVar (Var "text")) (expVar (Var "k"))) (expChr 'b'))
	                 (cmdGoto (Lbl "L1"))
	     ; cmdAssign (expVar (Var "k")) (expAdd (expVar (Var "k")) (expInt 1))
	     ; cmdGoto (Lbl "L6")
	     ]
    ; defLbl (Lbl "L6")
             [ cmdIfThen (expNotEq (expIndex (expVar (Var "text")) (expVar (Var "k"))) (expChr 'c'))
	                 (cmdGoto (Lbl "L1"))
	     ; cmdAssign (expVar (Var "k")) (expAdd (expVar (Var "k")) (expInt 1))
	     ; cmdGoto (Lbl "L7")
	     ]
    ; defLbl (Lbl "L7")
             [ cmdIfThen (expNotEq (expIndex (expVar (Var "text")) (expVar (Var "k"))) (expChr 'a'))
	                 (cmdGoto (Lbl "L0"))
	     ; cmdAssign (expVar (Var "k")) (expAdd (expVar (Var "k")) (expInt 1))
	     ; cmdGoto (Lbl "L8")
	     ]
    ; defLbl (Lbl "L8")
             [ cmdIfThen (expNotEq (expIndex (expVar (Var "text")) (expVar (Var "k"))) (expChr 'c'))
	                 (cmdGoto (Lbl "L5"))
	     ; cmdAssign (expVar (Var "k")) (expAdd (expVar (Var "k")) (expInt 1))
	     ; cmdGoto (Lbl "L9")
	     ]
    ; defLbl (Lbl "L9")
             [ cmdIfThen (expNotEq (expIndex (expVar (Var "text")) (expVar (Var "k"))) (expChr 'a'))
	                 (cmdGoto (Lbl "L0"))
	     ; cmdAssign (expVar (Var "k")) (expAdd (expVar (Var "k")) (expInt 1))
	     ; cmdGoto (Lbl "L10")
	     ]
    ; defLbl (Lbl "L10")
             [ cmdIfThen (expNotEq (expIndex (expVar (Var "text")) (expVar (Var "k"))) (expChr 'b'))
	                 (cmdGoto (Lbl "L1"))
	     ; cmdAssign (expVar (Var "k")) (expAdd (expVar (Var "k")) (expInt 1))
	     ; cmdReturn (expSub (expVar (Var "k")) (expInt 10))
	     ]
    ]
    [ cmdAssign (expVar (Var "k")) (expVar (Var "0"))
    ; cmdAssign (expVar (Var "n")) (expLength (expVar (Var "text")))
    ; cmdAssign (expVar (Var "text")) (expStrConcat (expVar (Var "text")) (expStr "@a"))
    ; cmdGoto (Lbl "L0")
    ]
]
[ cmdPrint (expCall (Var "kmp") [expStr "ababcabcabcacabcacab"]) ]

let () = print_string (unPrg kmp)
