(*                         J is for JavaScript:
 *                    A direct-style correspondence
 *             between Algol-like languages and JavaScript
 *                   using first-class continuations
 *
 *            Olivier Danvy, Chung-chieh Shan, and Ian Zerny
 *                            Copyright 2009
 *                  GPLv3, see LICENSE.txt for details
 *
 *                    www.zerny.dk/j-is-for-js.html
 *
 *                         Version: April 2011
 *)

#use "gen.ml";;

(* http://en.wikipedia.org/wiki/Man_or_boy_test
begin
  real procedure A (k, x1, x2, x3, x4, x5);
  value k; integer k;
  begin
    real procedure B;
    begin k:= k - 1;
          B:= A := A (k, B, x1, x2, x3, x4);
    end;
    if k <= 0 then A:= x4 + x5 else B;
  end;
  outreal (A (10, 1, -1, -1, 1, 0));
end;
*)

(* In ML -- http://lambda-the-ultimate.org/node/1309#comment-14814
local
   fun A (k, x1, x2, x3, x4, x5) =
       let
          fun B () =
              (k := !k - 1
             ; A (ref (!k), B, x1, x2, x3, x4))
       in
          if !k <= 0 then x4 () + x5 () else B ()
       end
in
   val () = print (Int.toString
                      (A (ref 10,
                          fn () => 1,
                          fn () => ~1,
                          fn () => ~1,
                          fn () => 1,
                          fn () => 0))^"\n")
end
*)

let manboy
= let a = Var "A" in
  let b = Var "B" in
  let k = Var "k" in
  let x1 = Var "x1" in
  let x2 = Var "x2" in
  let x3 = Var "x3" in
  let x4 = Var "x4" in
  let x5 = Var "x5" in
  program
    [ defFun x1 [] [] [cmdReturn (expInt 1)]
    ; defFun x2 [] [] [cmdReturn (expInt (-1))]
    ; defFun x3 [] [] [cmdReturn (expInt (-1))]
    ; defFun x4 [] [] [cmdReturn (expInt 1)]
    ; defFun x5 [] [] [cmdReturn (expInt 0)]
    ; defFun a [k;x1;x2;x3;x4;x5]
      [ defFun b [] [] [ cmdAssign (expVar k) (expSub (expVar k) (expInt 1))
		       ; cmdReturn (expCall a [ expVar k
					      ; expVar b
					      ; expVar x1
					      ; expVar x2
					      ; expVar x3
					      ; expVar x4 ]) ]
      ]
      [ cmdIfThen (expLt (expVar k) (expInt 1))
	  (cmdReturn (expAdd (expCall x4 []) (expCall x5 [])))
      ; cmdReturn (expCall b [])
      ]
    ]
    [ cmdPrint (expCall a [ expInt 10
			  ; expVar x1
			  ; expVar x2
			  ; expVar x3
			  ; expVar x4
			  ; expVar x5 ]) ]

let () = print_string (unPrg manboy)
