(*                         J is for JavaScript:
 *                    A direct-style correspondence
 *             between Algol-like languages and JavaScript
 *                   using first-class continuations
 *
 *            Olivier Danvy, Chung-chieh Shan, and Ian Zerny
 *                            Copyright 2009
 *                  GPLv3, see LICENSE.txt for details
 *
 *                    www.zerny.dk/j-is-for-js.html
 *
 *                         Version: April 2011
 *)

#use "gen.ml";;

(* Generate an invocation of a given quicksort *)
let quicksort qs_def
= let qs = Var "qs" in
  let a  = Var "A"  in
  let a_ var = expIndex (expVar a) (expVar var) in
  program
  [ defVar a ; qs_def qs a a_ ]
  [ cmdAssign (expVar a) (expArray (List.map expInt [5;2;1;4;6;3]))
  ; cmdPrint (expAdd (expStr "Random: ") (expVar a))
  ; cmdExp (expCall qs [expInt 0; expSub (expLength (expVar a)) (expInt 1)])
  ; cmdPrint (expAdd (expStr "Sorted: ") (expVar a)) ]

(* Quicksorting with gotos (Knuth, 1974, p. 285) *)
let qs_7a qs a a_
= let m  = Var "m"  in
  let n  = Var "n"  in
  let i  = Var "i"  in
  let j  = Var "j"  in
  let v  = Var "v"  in
  let loop1  = Lbl "loop1"  in
  let upt    = Lbl "upt"    in
  let loop2  = Lbl "loop2"  in
  let upf    = Lbl "upf"    in
  let common = Lbl "common" in
  defFun qs [m; n]
    [ defVar i
    ; defVar j
    ; defVar v
    ; defLbl loop1
      [ cmdIfThen (expGt (a_ i) (expVar v))
                  (cmdBlock [] [ cmdAssign (a_ j) (a_ i)
                               ; cmdGoto upf ])
      ; cmdGoto upt ]
    ; defLbl upt
      [ cmdAssign (expVar i) (expAdd (expVar i) (expInt 1))
      ; cmdIfThen (expLt (expVar i) (expVar j))
                  (cmdGoto loop1)
      ; cmdGoto common ]
    ; defLbl loop2
      [ cmdIfThen (expGt (expVar v) (a_ j))
                  (cmdBlock [] [ cmdAssign (a_ i) (a_ j)
                               ; cmdGoto upt ])
      ; cmdGoto upf ]
    ; defLbl upf
      [ cmdAssign (expVar j) (expSub (expVar j) (expInt 1))
      ; cmdIfThen (expLt (expVar i) (expVar j))
                  (cmdGoto loop2)
      ; cmdGoto common ]
    ; defLbl common
      [ cmdAssign (a_ j)
                  (expVar v)
      ; cmdIfThen (expGt (expSub (expVar n) (expVar m)) (expInt 1))
                  (cmdBlock [] [ cmdExp (expCall qs [expVar m; expSub (expVar j) (expInt 1)])
                               ; cmdExp (expCall qs [expAdd (expVar j) (expInt 1); expVar n]) ]) ] ]
    [ cmdAssign (expVar i) (expVar m)
    ; cmdAssign (expVar j) (expVar n)
    ; cmdAssign (expVar v) (a_ j)
    ; cmdGoto loop1 ]

(* Quicksorting with coroutines (Knuth, 1974, p. 287) *)
let qs_7b qs a a_
= let m  = Var "m"  in
  let n  = Var "n"  in
  let i  = Var "i"  in
  let j  = Var "j"  in
  let v  = Var "v"  in
  let loop   = Lbl "loop"   in
  let upt    = Lbl "upt"    in
  let upf    = Lbl "upf"    in
  let common = Lbl "common" in
  defFun qs [m; n]
    [ defVar i
    ; defVar j
    ; defVar v
    ; defLbl upt
      [ cmdBlock
        [ defLbl loop
          [ cmdIfThen (expGt (a_ i) (expVar v))
                      (cmdBlock [] [ cmdAssign (a_ j) (a_ i)
                                   ; cmdYield upt upf ])
          ; cmdAssign (expVar i) (expAdd (expVar i) (expInt 1))
          ; cmdIfThen (expLt (expVar i) (expVar j))
                      (cmdGoto loop)
          ; cmdGoto common ] ]
        [ cmdGoto loop ] ]
    ; defLbl upf
      [ cmdBlock
        [ defLbl loop
          [ cmdIfThen (expGt (expVar v) (a_ j))
                      (cmdBlock [] [ cmdAssign (a_ i) (a_ j)
                                   ; cmdYield upf upt ])
          ; cmdAssign (expVar j) (expSub (expVar j) (expInt 1))
          ; cmdIfThen (expLt (expVar i) (expVar j))
                      (cmdGoto loop)
          ; cmdGoto common ] ]
        [ cmdGoto loop ] ]
    ; defLbl common
      [ cmdAssign (a_ j)
                  (expVar v)
      ; cmdIfThen (expGt (expSub (expVar n) (expVar m)) (expInt 1))
                  (cmdBlock [] [ cmdExp (expCall qs [expVar m; expSub (expVar j) (expInt 1)])
                               ; cmdExp (expCall qs [expAdd (expVar j) (expInt 1); expVar n]) ]) ] ]
    [ cmdAssign (expVar i) (expVar m)
    ; cmdAssign (expVar j) (expVar n)
    ; cmdAssign (expVar v) (a_ j)
    ; cmdGoto upt ]

let () = print_string (unPrg (quicksort qs_7a))
let () = print_string (unPrg (quicksort qs_7b))
