(*                         J is for JavaScript:
 *                    A direct-style correspondence
 *             between Algol-like languages and JavaScript
 *                   using first-class continuations
 *
 *            Olivier Danvy, Chung-chieh Shan, and Ian Zerny
 *                            Copyright 2009
 *                  GPLv3, see LICENSE.txt for details
 *
 *                    www.zerny.dk/j-is-for-js.html
 *
 *                         Version: April 2011
 *)

#use "gen.ml";;

(*

  A tree is represented as an array, as we did for calder mobiles,
  only now we use -1 to denote a node and all leafs are
  integers.

procedure traverse (t, f) = 
  procedure visit (i) =
    n = t[i];
    if n != -1
    then return f(n)
    i := 2*i;
    visit (i+1);
    visit (i+2);
  end;
  visit (0)
end

procedure samefringe t1 t2 =
  block v1, v2 in

    procedure next1 (e) =
       v1 := e
       yield from l1 to l2
    end

    procedure next2 (e) =
       v2 := e
       yield from l2 to compare
    end

    l1: traverse (t1, next1); next1(undefined);
    l2: traverse (t2, next2); next2(undefined);
    compare:
      if v1 != v2 then return false
      if v1 = undefined then return true;
      goto l1
  end
end

t1 := [-1,1,-1,_,_,2,3];
t2 := [-1,-1,3,1,2,_,_];
print(samefringe(t1,t2));

*)

let samefringe = 
  let traverse = Var "traverse" in
  let samefringe = Var "samefringe" in
  let t1 = Var "t1" in
  let t2 = Var "t2" in
    program
      [ defVar t1
      ; defVar t2
      ; (let f = Var "f" in
	 let t = Var "t" in
	 let t_ v = expIndex (expVar t) (expVar v) in
         let visit = Var "visit" in
	 defFun traverse [t; f]
	   [ let i = Var "i" in
  	     let n = Var "n" in
	     defFun visit [i]
	       [ defVar n ]
	       [ cmdAssign (expVar n) (t_ i)
	       ; cmdIfThen (expNotEq (expVar n) (expInt (-1)))
		   (cmdReturn (expCall f [expVar n]))
	       ; cmdAssign (expVar i) (expMul (expInt 2) (expVar i))
	       ; cmdExp (expCall visit [expAdd (expVar i) (expInt 1)])
	       ; cmdExp (expCall visit [expAdd (expVar i) (expInt 2)])
	       ]
	   ]
	   [ cmdExp (expCall visit [expInt 0])])
      ; let t1 = Var "t1" in
        let t2 = Var "t2" in
	let v1 = Var "v1" in
	let v2 = Var "v2" in
	let e  = Var "e" in
	let next1 = Var "next1" in
	let next2 = Var "next2" in
	let l1 = Lbl "l1" in
	let l2 = Lbl "l2" in
	let compare = Lbl "compare" in
	defFun samefringe [t1; t2]
	  [ defVar v1
	  ; defVar v2
	  ; defFun next1 [e] []
	    [ cmdAssign (expVar v1) (expVar e)
	    ; cmdYield l1 l2
	    ]
	  ; defFun next2 [e] []
	    [ cmdAssign (expVar v2) (expVar e)
	    ; cmdYield l2 compare
	    ]
	  ; defLbl l1 [ cmdExp (expCall traverse [expVar t1; expVar next1])
		      ; cmdExp (expCall next1 [expBottom]) ]
	  ; defLbl l2 [ cmdExp (expCall traverse [expVar t2; expVar next2])
		      ; cmdExp (expCall next2 [expBottom]) ]
	  ; defLbl compare
	    [ cmdIfThen (expNotEq (expVar v1) (expVar v2))
		(cmdReturn (expBool false))
	    ; cmdIfThen (expEq (expVar v1) expBottom)
	        (cmdReturn (expBool true))
	    ; cmdGoto l1
	    ]
	  ]
	  [ cmdGoto l1 ]
      ]
      [ cmdAssign (expVar t1)
	  (expArray [ expInt (-1)
		    ; expInt 1; expInt (-1)
		    ; expBottom; expBottom; expInt 2; expInt 3])
      ; cmdAssign (expVar t2)
	  (expArray [ expInt (-1)
		    ; expInt (-1); expInt 3
		    ; expInt 1; expInt 2; expBottom; expBottom])
      ; cmdPrint (expCall samefringe [expVar t1; expVar t2])
      ]

let () = print_string (unPrg samefringe)
